#include <PalmOS.h>

#include "plugin.h"
#include "db.h"
#include "enum-memoexport.h"

#define memoDBType						'DATA'

typedef struct {
	char 	note;		// null terminated
	UInt8	reserved;
} MemoDBRecordType;

typedef MemoDBRecordType * MemoDBRecordPtr;

Err MemoNewRecord (DmOpenRef dbP, Char *data, UInt32  size)
{
    UInt16              index;
	UInt32				offset;
	MemHandle			recordH;	
	MemoDBRecordPtr	recordP, nilP=0;

    index = DmNumRecords (dbP);
	//  Allocate a chunk in the database for the new record.
	recordH = DmNewRecord(dbP, &index, size + 1);
	if (recordH == NULL)
		return dmErrMemError;

	// Pack the the data into the new record.
	recordP = MemHandleLock (recordH);
	offset = (UInt32)&nilP->note;
	DmStrCopy(recordP, offset, data);
	
	MemPtrUnlock (recordP);

	DmReleaseRecord (dbP, index, true);
	return errNone;
}

UInt32 FillData(DataSource * source, UInt16 record, Char ** data)
{
    UInt32 size = 0;
    UInt16 i;
    DataSourceGetter getter;
    void * row_data;
    Char string[STRING_MAXCHAR];
    Char *tempo;
    
    source->ops.LockRecord(source, record, false, &getter, &row_data);

    for (i = 0; i < source->schema.numFields; i++) {
        size += StrLen(source->schema.fields[i].name) + 2;

        ConvertFieldToString( source, &getter, row_data,
                        i, string, STRING_MAXCHAR);
        size += StrLen(string) + 1;
    }
    size += 1;
    tempo = *data = MemPtrNew(size);

    for (i = 0; i < source->schema.numFields; i++) {
        ConvertFieldToString( source, &getter, row_data,
                        i, string, STRING_MAXCHAR);

        StrPrintF(tempo, "%s: %s\n", source->schema.fields[i].name, string);
        tempo += StrLen(source->schema.fields[i].name) + 2 + StrLen(string) + 1;
    }

    (*data)[size -1] = 0;
    source->ops.UnlockRecord(source, record, false, row_data);
    return size;
}

void MyCallBack( void *plugData)
{
    Char *data;
    UInt32 size;
    DmOpenRef db;
    
    if (FrmAlert(alertID_ConfirmCreation))
        return;

    size = FillData(Plug_GetSource(plugData), Plug_GetRecord(plugData), &data);

    db = DmOpenDatabaseByTypeCreator (memoDBType, sysFileCMemo, dmModeReadWrite);
    if (db) {
        MemoNewRecord(db, data, size);
        DmCloseDatabase(db);
    }
    MemPtrFree(data);
}

void MainPlug(void *plugData)
{
    Plug_SetDriver(plugData, NULL);
    Plug_SetAboutAlert(plugData, alertID_AboutBox);
    Plug_SetAction(plugData, ActionRunInEditViewSelector, MyCallBack);
}
