/*
 * DB: Report Popup
 * 
 *   - A minor variation of the Note form (note.c)
 *
 * Copyright (C) 2001-2003: Marc Chalain (marc-chalain@voila.fr)
 * Copyright (C) 2003: Scott Wallace (swallace@umich.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <PalmOS.h>
#include <Window.h>
#include <unix_stdarg.h>

#include "db.h"
#include "enum-report.h"

#include "util.h"
#include "display.h"


extern void printReport( DataSource *source, UInt16 field, char *buff );
#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))

static void DisplayUpdateScrollers() EDITSECT;
static void DisplayScroll (WinDirectionType direction, UInt16 linesToScroll) EDITSECT;
static Boolean DisplayHandleEvent(EventPtr event);

UInt16 DisplayPopupForm( PersistentData *pd)
{
	FormPtr noteForm, oldForm;
	FieldPtr fld;
	FieldAttrType attr;
	UInt16 retVal = noteUnchanged;
	ListPtr list;


	oldForm = FrmGetActiveForm();
	noteForm = FrmInitForm(formID_Report);
	FrmSetActiveForm(noteForm);
	FrmSetEventHandler(noteForm, DisplayHandleEvent);

	/**
	 * Initialize the form 
	 */
	fld = GetObjectPtr(noteForm, ctlID_Report_Display);
	
    FldSetTextHandle(fld, pd->note.h);
    DisplayUpdateScrollers();
	
	if (pd->note.title[0]) {
		FrmSetTitle (noteForm, pd->note.title);
	} 
	
	FldSetMaxChars(fld, NOTE_MAXCHAR - 1);
	FldGetAttributes( fld, &attr);
	attr.editable = false;
//	attr.underlined = noUnderline;

	attr.hasScrollBar = true;
	attr.autoShift = true;
	FldSetAttributes( fld, &attr);

	list = GetObjectPtr( noteForm, ctlID_Report_FieldList );
	LstSetListChoices( list, pd->fields, pd->n );
	LstSetHeight( list, (pd->n>10)?10:pd->n );
	LstSetSelection( list, noListSelection );

	// This is not in the spirit of the Gadget!
	FrmSetGadgetData(noteForm, FrmGetObjectIndex(noteForm,
											 ctlID_Report_PersistentData), pd);
	
	FrmDrawForm(noteForm);
	
	/* Handle events until the button is pressed */
	FrmDoDialog(noteForm);


	FldSetTextHandle(fld, 0);


	FrmDeleteForm(noteForm);
	FrmSetActiveForm(oldForm);

	/* I think this was required to get rid of some weird artifacts
	 *   when returning to the edit view, but it does not seem to be
	 * required when returning to the list view.  
	 */
//	FrmUpdateForm(FrmGetFormId(oldForm), updateRedrawForm);


    return retVal;
}


/***********************************************************************
 *
 * FUNCTION:    DisplayUpdateScrollers
 *
 * DESCRIPTION: P13. Visually updates the scroll arrows.
 *
 * PARAMETERS:  frm - poInt16er to the Edit View form.
 *
 * RETURNED:    nothing
 *
 ***********************************************************************/
static void DisplayUpdateScrollers()
{
	FieldPtr fieldP;
	ScrollBarPtr scrollBarP;

	UInt16 nPosition, nMax;
	UInt16 textHeight, fieldHeight;
    FormPtr form = FrmGetActiveForm();

	fieldP = GetObjectPtr(form,ctlID_Report_Display);
	scrollBarP = GetObjectPtr(form,ctlID_Report_ScrollBar);

	FldGetScrollValues(fieldP, &nPosition, &textHeight, &fieldHeight);

	if (textHeight > fieldHeight){
		nMax = textHeight - fieldHeight + FldGetNumberOfBlankLines (fieldP);
	} else if (nPosition) {
		nMax = nPosition;
	} else {
		nMax = 0;
	}

	SclSetScrollBar(scrollBarP, nPosition, 0, nMax, fieldHeight - 1);

}



/***********************************************************************
 *
 * FUNCTION:    DisplayScroll
 *
 * DESCRIPTION: P13. Scrolls the edit view a page or a line at a time.
 *
 * PARAMETERS:  direction - up or dowm
 *              oneLine   - true scroll one line, false scrolls one page
 *
 * RETURNED:    nothing
 *
 ***********************************************************************/
static void DisplayScroll (WinDirectionType direction, UInt16 linesToScroll)
{
	FieldPtr fld;
	Int16 blankLines;

	fld = GetObjectPtr(FrmGetActiveForm(),ctlID_Report_Display);
	if (FldScrollable (fld, direction))
	{
		blankLines = FldGetNumberOfBlankLines(fld);

		FldScrollField(fld, linesToScroll, direction);
		DisplayUpdateScrollers();
	}
}



/***********************************************************************
 *
 * FUNCTION:		DisplayHandleEvent
 *
 * DESCRIPTION:	P3. Handles processing of events for the edit form.
 *
 * PARAMETERS:		event	- the most recent event.
 *
 * RETURNED:		True if the event is handled, false otherwise.
 *
 ***********************************************************************/
static Boolean DisplayHandleEvent(EventPtr event)
{
	FormPtr		frm;
	FieldPtr	fld;
	Boolean	handled = false;

	switch (event->eType)
	{

// No need for open/close events. This form is entered with DoDialog
    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_Report_FieldTrigger:
		{
			
			PersistentData *pd;
			FieldPtr fld;
			ListPtr lst;
			ControlPtr ctl;
			int sel;

			frm = FrmGetActiveForm();
			pd = (PersistentData *)FrmGetGadgetData( frm, FrmGetObjectIndex( frm, 
											  ctlID_Report_PersistentData ));
			
			lst = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, ctlID_Report_FieldList ));
			
			sel = LstPopupList( lst );
			if ( sel != noListSelection ) {
				fld = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm,
														   ctlID_Report_Display ));

				
				FldSetTextHandle( fld, 0 );
				printReport( pd->source, sel, MemHandleLock( pd->note.h ) );
				MemHandleUnlock( pd->note.h );
				
				FldSetTextHandle( fld, pd->note.h );
				FldDrawField( fld );
				FldSendChangeNotification( fld );
				ctl = FrmGetObjectPtr( frm, FrmGetObjectIndex( frm,
												   ctlID_Report_FieldTrigger));
				CtlSetLabel( ctl, pd->fields[sel] );
				CtlDrawControl( ctl );
			}
			
		}

	}
		
		break;

   	case keyDownEvent:
   		if (event->data.keyDown.chr == pageUpChr)
		{
			UInt16 fieldHeight, null1, null2;
			fld = GetObjectPtr(FrmGetActiveForm(),ctlID_Report_Display);
			FldGetScrollValues(fld, &null1, &null2, &fieldHeight);
			DisplayScroll (winUp, fieldHeight);
			handled = true;
		}
	
		else if (event->data.keyDown.chr == pageDownChr)
		{
			UInt16 fieldHeight, null1, null2;
			fld = GetObjectPtr(FrmGetActiveForm(),ctlID_Report_Display);
			FldGetScrollValues(fld, &null1, &null2, &fieldHeight);
			DisplayScroll (winDown, fieldHeight);
			handled = true;
		}
		else
		{
			frm = FrmGetActiveForm ();
			FrmHandleEvent (frm, event);
			DisplayUpdateScrollers ();
			handled = true;
		}
		break;
			
	case sclRepeatEvent:
		if( (event->data.sclRepeat.newValue - event->data.sclRepeat.value) < 0)
			DisplayScroll (winUp, event->data.sclRepeat.value - event->data.sclRepeat.newValue);
		else
	   		DisplayScroll (winDown, event->data.sclRepeat.newValue - event->data.sclRepeat.value);
		break;
			
			
    case fldChangedEvent:
        DisplayUpdateScrollers();
    break;


    default:
    break;
    }
    return(handled);
}

/*
void
DbgPrintF(const Char * fmt, ...)
{
    va_list args;
    Char buf[256];

#define DEBUG_OUT_NAME "report-debug.txt"
	HostFILE  *hf=NULL;

    va_start(args, fmt);
    HostTraceOutputVTL( 0, fmt, args);
    hf = HostFOpen(DEBUG_OUT_NAME,"a");
    if (hf) {
        StrVPrintF(buf, fmt, args);
	    HostFPrintF(hf,buf);
	    HostFPrintF(hf,"\n");
	    HostFClose(hf);
    }
    va_end(args);

}
*/

#undef GetObjectPtr
