/*
 * DB: Memory debugging tools
 * Copyright (C) 2002-2003: Scott Wallace (swallace@mynah.eecs.umich.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include "leakdetect.h"

#ifdef DETECT_LEAKS
#define LEAK_OUT_NAME "leaks.txt"
#define LEAK_OUT_ROOT "leaks"

void RecordMemInfoFn( void *addr, char *str, char *file, long line ) {
	HostFILE  *hf=NULL;

	ASSERT(file);
	
	hf = HostFOpen(LEAK_OUT_NAME,"a");
	if (hf) {
		HostFPrintF(hf,"i 0x%lx '%s' %s %ld\n", (long)addr, str, file, line);
		HostFClose(hf);
#ifdef CONFIG_DEBUG
        HostTraceOutputTL( 0, "i 0x%lx '%s' %s %ld\n", (long)addr, str, file, line);
#endif
	}
}


MemHandle AllocateMemHandleFn(long size, char *file, long line)
{
     MemHandle   addr=NULL;
     HostFILE  *hf=NULL;

     ASSERT(file);

    if (!size) return NULL;
     addr = MemHandleNew(size);
     hf = HostFOpen(LEAK_OUT_NAME,"a");
     if (hf) {
	  HostFPrintF(hf,"ha 0x%lx 0x%lx %s %ld\n", (long)addr, (long)size, file, line);
	  HostFClose(hf);
#ifdef CONFIG_DEBUG
      HostTraceOutputTL( 0, "ha 0x%lx 0x%lx %s %ld\n", (long)addr, (long)size, file, line);
#endif
     }
     return addr;
}

Err DeallocateMemHandleFn(MemHandle addr, char *file, long line)
{
     HostFILE  *hf=NULL;
	 Err e;

     ASSERT(addr);
     ASSERT(file);

	 if ( ((UInt32)addr) % 2 ) {
		 // open, append, and close because we're about to crash...
		 hf = HostFOpen(LEAK_OUT_NAME,"a");

		 if (hf) {
			 HostFPrintF(hf,"h!f 0x%lx %s %ld\n", (long) addr, file, line);
			 HostFPrintF(hf,"i 0x%lx 'Internal Msg: Handle is not aligned, about to crash' %s %ld\n", (long) addr, file, line);
			 HostFClose(hf);
			 hf = NULL;

#ifdef CONFIG_DEBUG
			 HostTraceOutputTL( 0,"h!f 0x%lx %s %ld\n", (long) addr, file, line);
			 HostTraceOutputTL( 0,"i 0x%lx 'Internal Msg: Handle is not aligned, about to crash' %s %ld\n", (long) addr, file, line);
#endif

		 }
		 
	 }
	 
     e = MemHandleFree(addr);
     hf = HostFOpen(LEAK_OUT_NAME,"a");

     if (hf) {
		 if ( !e ) {
			 HostFPrintF(hf,"hf 0x%lx %s %ld\n", (long) addr, file, line);
#ifdef CONFIG_DEBUG
			 HostTraceOutputTL( 0, "hf 0x%lx %s %ld\n", (long) addr, file, line);
#endif
		 }
		 else {
			 HostFPrintF(hf,"h!f 0x%lx %s %ld\n", (long) addr, file, line);
			 HostFPrintF(hf,"i 0x%lx 'Internal Msg: Error Code %lu' %s %ld\n", (long) addr, (UInt32)e,  file, line);

#ifdef CONFIG_DEBUG
			 HostTraceOutputTL( 0, "h!f 0x%lx %s %ld\n", (long) addr, file, line);
			 HostTraceOutputTL( 0,"i 0x%lx 'Internal Msg: Error Code %lu' %s %ld\n", (long) addr, (UInt32)e,  file, line);
#endif


		 }
		 HostFClose(hf);
     }
	 return e;


}

MemPtr AllocateMemPtrFn(long size, char *file, long line)
{
     MemPtr   addr=NULL;
     HostFILE  *hf=NULL;

     ASSERT(file);

    if (!size) return NULL;
     addr = MemPtrNew(size);
     hf = HostFOpen(LEAK_OUT_NAME,"a");
     if (hf) {
	  HostFPrintF(hf,"pa 0x%lx 0x%lx %s %ld\n", (long)addr, (long)size, file, line);
	  HostFClose(hf);
#ifdef CONFIG_DEBUG
	  HostTraceOutputTL(0,"pa 0x%lx 0x%lx %s %ld\n", (long)addr, (long)size, file, line);
#endif
     }
     return addr;
}

Err DeallocateMemPtrFn(MemPtr addr, char *file, long line)
{
     HostFILE  *hf=NULL;
	 Err e;

     ASSERT(addr);
     ASSERT(file);

    if (!addr) {
     hf = HostFOpen(LEAK_OUT_NAME,"a");
     if (hf) {
	  HostFPrintF(hf,"NF 0x%lx %s %ld\n", (long) addr, file, line);
	  HostFClose(hf);
#ifdef CONFIG_DEBUG
	  HostTraceOutputTL(0,"NFf 0x%lx %s %ld\n", (long) addr, file, line);
#endif
     }
    }

     e = MemPtrFree(addr);
     hf = HostFOpen(LEAK_OUT_NAME,"a");
     if (hf) {
		 if (!e) {
			 HostFPrintF(hf,"pf 0x%lx %s %ld\n", (long) addr, file, line);

#ifdef CONFIG_DEBUG
			 HostTraceOutputTL(0,"pf 0x%lx %s %ld\n", (long) addr, file, line);
#endif
		 }
		 else {
			 HostFPrintF(hf,"p!f 0x%lx %s %ld\n", (long) addr, file, line);
			 HostFPrintF(hf,"i 0x%lx 'Internal Msg: Error Code %lu' %s %ld\n", (long) addr, (UInt32)e,  file, line);

#ifdef CONFIG_DEBUG
			 HostTraceOutputTL(0,"p!f 0x%lx %s %ld\n", (long) addr, file, line);
#endif
		 }
		 HostFClose(hf);

     }
	 return e;

}


MemPtr LockMemHandleFn(MemHandle addr, char *file, long line)
{
    HostFILE  *hf=NULL;
    MemPtr p;

    ASSERT(addr);
    ASSERT(file);

    if (!addr) {
        hf = HostFOpen(LEAK_OUT_NAME,"a");
        if (hf) {
            HostFPrintF(hf,"NF 0x%lx %s %ld\n", (long) addr, file, line);
            HostFClose(hf);
#ifdef CONFIG_DEBUG
            HostTraceOutputTL(0,"NFl 0x%lx %s %ld\n", (long) addr, file, line);
#endif
        }
    }

    p = MemHandleLock(addr);
    hf = HostFOpen(LEAK_OUT_NAME,"a");
    if (hf) {
        if ( p )
            HostFPrintF(hf,"hl 0x%lx %s %ld\n", (long) addr, file, line);
        HostFClose(hf);
#ifdef CONFIG_DEBUG
        HostTraceOutputTL(0,"hl 0x%lx %s %ld\n", (long) addr, file, line);
#endif
    }
    return p;    
}

Err UnlockMemHandleFn(MemHandle addr, char *file, long line)
{
    HostFILE  *hf=NULL;
    Err e;

    ASSERT(addr);
    ASSERT(file);

    if (!addr) {
        hf = HostFOpen(LEAK_OUT_NAME,"a");
        if (hf) {
            HostFPrintF(hf,"NF 0x%lx %s %ld\n", (long) addr, file, line);
            HostFClose(hf);
#ifdef CONFIG_DEBUG
            HostTraceOutputTL(0,"NFu 0x%lx %s %ld\n", (long) addr, file, line);
#endif
        }
    }

	if ( ((UInt32)addr) % 2 ) {
		hf = HostFOpen(LEAK_OUT_NAME,"a");	
        if (hf) {
            HostFPrintF(hf,"h!u 0x%lx %s %ld\n", (long) addr, file, line);
			HostFPrintF(hf,"i 0x%lx 'Internal Msg: Handle is not aligned, about to crash' %s %ld\n", (long) addr, file, line);

			HostFClose(hf);
			hf = NULL;

#ifdef CONFIG_DEBUG
			 HostTraceOutputTL( 0,"h!u 0x%lx %s %ld\n", (long) addr, file, line);
			 HostTraceOutputTL( 0,"i 0x%lx 'Internal Msg: Handle is not aligned, about to crash' %s %ld\n", (long) addr, file, line);
#endif

		}
		
	}

	hf = HostFOpen(LEAK_OUT_NAME,"a");	
    e = MemHandleUnlock(addr);

    if (hf) {
		if (!e) {
			HostFPrintF(hf,"hu 0x%lx %s %ld\n", (long) addr, file, line);

#ifdef CONFIG_DEBUG
			HostTraceOutputTL(0,"hu 0x%lx %s %ld\n", (long) addr, file, line);
#endif
		}
		else {
			HostFPrintF(hf,"h!u 0x%lx %s %ld\n", (long) addr, file, line);
			HostFPrintF(hf,"i 0x%lx 'Internal Msg: Error Code %lu' %s %ld\n", (long) addr, (UInt32)e,  file, line);

#ifdef CONFIG_DEBUG
			HostTraceOutputTL(0,"h!u 0x%lx %s %ld\n", (long) addr, file, line);
			HostTraceOutputTL( 0,"i 0x%lx 'Internal Msg: Error Code %lu' %s %ld\n", (long) addr, (UInt32)e,  file, line);

#endif
		}
		HostFClose(hf);
    }

    return e;    
}


Err UnlockMemPtrFn(MemPtr addr, char *file, long line)
{
    HostFILE  *hf=NULL;
    Err e;
    MemHandle hand;

    ASSERT(addr);
    ASSERT(file);


	if ( ((UInt32)addr) % 2 ) {
        hf = HostFOpen(LEAK_OUT_NAME,"a");
        if (hf) {
            HostFPrintF(hf,"h!u 0x%lx %s %ld\n", (long) addr, file, line);
			HostFPrintF(hf,"i 0x%lx 'Internal Msg: Pointer is not aligned, address refers to pointer...wont be able to recover handle... about to crash' %s %ld\n", (long) addr, file, line);

			HostFClose(hf);
			hf = NULL;

#ifdef CONFIG_DEBUG
			 HostTraceOutputTL( 0,"h!f 0x%lx %s %ld\n", (long) addr, file, line);
			 HostTraceOutputTL( 0,"i 0x%lx 'Internal Msg: Handle is not aligned, about to crash' %s %ld\n", (long) addr, file, line);
			 HostTraceOutputTL( 0,"i 0x%lx 'Internal Msg: Pointer is not aligned, address refers to pointer...wont be able to recover handle... about to crash' %s %ld\n", (long) addr, file, line);

#endif

		}
		
	}

    hand = MemPtrRecoverHandle(addr);
    e = MemPtrUnlock(addr);
    hf = HostFOpen(LEAK_OUT_NAME,"a");
    if (hf) {
		if (!e) {
			HostFPrintF(hf,"hu 0x%lx %s %ld\n", (long) hand, file, line);

#ifdef CONFIG_DEBUG
			HostTraceOutputTL(0,"hu 0x%lx %s %ld\n", (long) hand, file, line);
#endif
		}
		else {
			HostFPrintF(hf,"h!u 0x%lx %s %ld\n", (long) hand, file, line);
			HostFPrintF(hf,"i 0x%lx 'Internal Msg: Error Code %lu' %s %ld\n", (long) addr, (UInt32)e,  file, line);


#ifdef CONFIG_DEBUG
			HostTraceOutputTL(0,"h!u 0x%lx %s %ld\n", (long) hand, file, line);
			HostTraceOutputTL( 0,"i 0x%lx 'Internal Msg: Error Code %lu' %s %ld\n", (long) addr, (UInt32)e,  file, line);

#endif
		}

		HostFClose(hf);			
    }

    return e;    
}

void LeakDetect_FinalCallBeforeStoppingApp() {
    HostFILE  *hf=NULL;
	char newName[120];
	UInt32 secs;

    hf = HostFOpen(LEAK_OUT_NAME,"a");
    if (hf) {
		HostFClose(hf);
		secs = TimGetSeconds();
		StrPrintF( newName, "%s-%lu.txt", LEAK_OUT_ROOT, secs );
		HostRename(LEAK_OUT_NAME, newName );
	}

}

#endif


