/*
 * DB: Record List View
 * Copyright (C) 1998-2001 by Tom Dyas (tdyas@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#define ALLOW_ACCESS_TO_INTERNALS_OF_CONTROLS

#include <PalmOS.h>
#include <PalmOSGlue.h>
#include <UIResources.h>

#ifdef CONFIG_HANDERA
#include <Vga.h>
#endif
#ifdef CONFIG_SONY
#include <SonyCLIE.h>
 #ifdef Capable
  #undef Capable
 #endif
#endif

#include "db.h"
#include "enum.h"
#include "grid.h"
#include "prefs.h"
#include "io.h"
#include "listview.h"
#include "util.h"
#include "popuplist.h"
#ifdef CONFIG_PLUGINS
#include "plug.h"
#endif
#ifdef CONFIG_GLB_SCRIPTS
#include "calcvalue.h"
#endif
#ifdef CONFIG_ICONS
#include "enum-icons.h"
#endif
#include "localsearch.h"
#ifdef CONFIG_REPORT
#include "report.h"
#include "note.h"
#endif

#define updateRunGlobalScript          0x1000
#define updateChangeViewWithTrigger (updateRunGlobalScript | updateChangeForm)



#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))
#define noRecord dmMaxRecordIndex

/*
 * if a database has more than this number of rows, the grid structure
 * will be retained in certain cases (e.g. when switching to the edit form)
 * this will often result in a time savings, but will also require more
 * dynamic memory.  In addition BE SURE that this size is ALWAYS GREATER than
 * the number of screen rows.  Otherwise, the grid might get messed up for
 * very small databases. 
 *
 * There is a strange bug I cannot figure out 
 *  (described later in this file).  As a work around I set this value
 *  unreasonably high
 */
#define AUTO_RETAIN_SIZE 1000000

static UInt16 ActiveViewIndex;
static DataSourceFeature_ListView * ActiveView;
static GridModelType * ViewModel;
static void *ViewModelData;
static Boolean RetainGrid = false;
static Char ** FilterMenuItems;

#ifdef CONFIG_PLUGINS
#ifdef CONFIG_ICONS
#include "plug.h"
static Char ** PlugInMenuItems;
#endif
#endif

GridType *Gridp;

extern FeatureType FeatureEditor_Feature;

static void UpdateFieldButtons(FormPtr, Boolean, Boolean) SECT_UI;
static void UpdateScrollers(FormPtr, Boolean, Boolean) SECT_UI;
static Boolean DoCommand(UInt16 itemID) SECT_UI;
static Boolean ListViewUpdateForm(FormPtr form, UInt16 type) SECT_UI;
static void SelectViewTrigger(void) SECT_UI;
static void DoQuickSort( UInt16 fieldNum , UInt16 sortOrder) SECT_UI;
static Boolean ListViewMove(WinDirectionType direction) SECT_UI;

//extern void InvalidateGrid(GridType *)GRIDSECT;

void ListViewFreeCachedData() {
	if ( RetainGrid ) {

		Gridp->ops->GridFree(Gridp);
//		RetainGrid = false;
	}
}

static void
UpdateFieldButtons(FormPtr form, Boolean scrollLeft, Boolean scrollRight)
{
    UInt16 leftIndex, rightIndex;

    leftIndex = FrmGetObjectIndex(form, ctlID_ListView_LeftButton);
    rightIndex = FrmGetObjectIndex(form, ctlID_ListView_RightButton);

    if (scrollLeft)
        FrmShowObject(form, leftIndex);
    else
        FrmHideObject(form, leftIndex);

    if (scrollRight)
        FrmShowObject(form, rightIndex);
    else
        FrmHideObject(form, rightIndex);
}

static void
UpdateScrollers(FormPtr form, Boolean scrollUp, Boolean scrollDown)
{
    UInt16 upIndex, downIndex;

    upIndex = FrmGetObjectIndex(form, ctlID_ListView_UpButton);
    downIndex = FrmGetObjectIndex(form, ctlID_ListView_DownButton);
    FrmUpdateScrollers(form, upIndex, downIndex, scrollUp, scrollDown);
}

static Boolean
DoCommand(UInt16 itemID)
{
    switch (itemID) {
    case menuitemID_DatabaseDesign:
        if ((CurrentSource->openMode != dmModeReadOnly)
                && (CurrentSource->driver->class_ops.Create))
        {

			// SW NOTE:  Is this really necessary? 
			// we should try to reduce the use of SetFilter
            if( CurrentSource->ops.Capable(CurrentSource, DS_CAPABLE_LISTVIEW_FILTER))
            {
                CurrentSource->ops.SetFilter( CurrentSource, dbNoFilter, true, NULL);
            }
            DesignViewDriver = CurrentSource->driver;
            FrmGotoForm(formID_DesignView);
        }
        return true;

    case menuitemID_Find:
		FindInfo.forFilterDefinition = -1;
        FrmPopupForm(formID_FindDialog);
        return true;

    case menuitemID_Repeat:
        if ((DBPrefs.flags & prefFlagUnuseFilters)
			|| !CurrentSource->ops.Capable(CurrentSource, DS_CAPABLE_LISTVIEW_FILTER))
        {
			UInt16 frec;

			if ( CurrentSource->ops.Search )
				frec = CurrentSource->ops.Search(CurrentSource, &FindInfo, dbNoFilter,
												 FILTER_UNION);
			else
				frec = DS_Search(CurrentSource, &FindInfo, dbNoFilter, FILTER_UNION);

				
			ListViewUpdateAfterFind( dbNoFilter, frec, &FindInfo );

		}
		return true;		

#ifdef CONFIG_SORT
    case menuitemID_Sort:
        if ( CurrentSource->ops.Sort && CurrentSource->openMode != dmModeReadOnly)
        {
            FrmPopupForm(formID_SortEditor);
        }
        return true;
#endif
	case menuitemID_RebuildRecords:
	{
        void * context;
		
		if ( CurrentSource->ops.RecalculateFields )
        {
			/* Display an on-screen "Working..." message. */
			context = ShowWorkingIndicator();
			CurrentSource->ops.RecalculateFields( CurrentSource );
			HideWorkingIndicator(context);
			FrmUpdateForm( formID_ListView, updateChangeForm );
		}
		return true;


	}
    case menuitemID_RebuildDatabase:
    {
        void * context;

        /* Display an on-screen "Working..." message. */
        context = ShowWorkingIndicator();
        CurrentSource = RebuildDatabase(CurrentSource, NULL, NULL);
        HideWorkingIndicator(context);
        ViewModel = ViewModel_New( CurrentSource, ActiveView, &ViewModelData);
        FrmUpdateForm( formID_ListView, updateChangeForm );
        return true;
    }
    case menuitemID_NewRecord:
        if (CurrentSource->openMode != dmModeReadOnly)
        {
            IsNewRecord = true;
			Gridp->numModelRows++;
			// we only keep the cache around for reasonably large databases
			// for small ones, (with number of records < screen size)
			// it is important to rebuild...
			// see the note in grid.c regarding grid->numModelRows
			if ( Gridp->numModelRows > AUTO_RETAIN_SIZE )
				RetainGrid = true;

            FrmGotoForm(formID_EditView);
        }
        return true;

    case menuitemID_DeleteAllRecords:
        if (CurrentSource->openMode != dmModeReadOnly 
                && FrmAlert(alertID_ConfirmDelete) == 0)
        {
            UInt16 recNum = 0;
            while (1)
            {
                if (! CurrentSource->ops.Seek(CurrentSource, &recNum, 0,
                                              dmSeekForward))
                {
                    break;
                }
                CurrentSource->ops.DeleteRecord(CurrentSource, recNum);
            }
            FrmUpdateForm(formID_ListView, updateChangeView);
        }
        return true;

    case menuitemID_GotoTop:
        CurrentRecord = 0;
        FrmUpdateForm(formID_ListView, updateRedrawForm);
        return true;

    case menuitemID_GotoBottom:
        CurrentRecord = dmMaxRecordIndex;
        FrmUpdateForm(formID_ListView, updateRedrawForm);
        return true;

    case menuitemID_DatabasePrefs:
        MenuEraseStatus(0);
        PopupDBPrefs();
        return true;

#ifdef CONFIG_GLB_SCRIPTS
	case menuitemID_ConfigGlobalScripts:
		FeatureEditor_Feature = featureGlobalScript;
		FrmGotoForm(formID_NamesEditor);
		return true;
#endif

    case menuitemID_Help:
        FrmHelp(stringID_Help_ListView);
        return true;

#ifdef CONFIG_PLUGINS
    case menuitemID_Plugins:
    {
        PlugEnvironement pluginInfo;
        pluginInfo.screen = PLUGINTYPE_LISTVIEW;
        pluginInfo.source = CurrentSource;
        Plug_ShowList( pluginInfo);
        return true;
    }
#endif
    default:
        return HandleCommonMenuEvent(itemID);

    }
}

static void
SelectViewTrigger(void)
{
    FormType * form;
    ListType * lst;
    UInt16 numViews, i;
    Int16 new_view_index;
    char ** view_names;

    form = FrmGetActiveForm();

    /* Construct an array of the names of the views in this database. */
	numViews = CurrentSource->ops.GetNumFeatures(CurrentSource, featureListView);
	view_names = AllocateMemPtr((numViews + 2) * sizeof(char *));
	for (i = 0; i < numViews; ++i)
    {
		DataSourceFeature_ListView * v = CurrentSource->ops.GetFeature(CurrentSource, featureListView, i);
		view_names[i] = AllocateMemPtr(1 + StrLen(v->name));
		StrCopy(view_names[i], v->name);
		CurrentSource->ops.PutFeature(CurrentSource, featureListView, i, v);
	}
    /* add "All" View and Edit list Views*/
    view_names[i++] = CopyStringResource(stringID_AllFields);
    view_names[i++] = CopyStringResource(stringID_EditListViews);

    /* Setup the list and popup trigger. */
    lst = GetObjectPtr(form, ctlID_ListView_ViewList);
    LstSetListChoices(lst, view_names, numViews + 2);
    LstSetHeight(lst, numViews + 2);
    LstSetSelection(lst, ActiveViewIndex);

    new_view_index = LstPopupList(lst);

    /* check if Edit list Views selected*/
    if (new_view_index == numViews + 1)
    {
		FeatureEditor_Feature = featureListView;
        FrmGotoForm(formID_NamesEditor);
    }
#ifdef CONFIG_GLB_SCRIPTS
    else if (new_view_index == ActiveViewIndex)
    {

        /* The global script might change the order of records, 
            * so we must call with updateChangeForm | updateRunGlobalScript
                    */
        FrmUpdateForm(formID_ListView, updateChangeViewWithTrigger);
    }
#endif
    /* If the user switched views, move to the new view. */
    /* new_view_index != ActiveViewIndex */
    /* select default view */
    else if (new_view_index == numViews)
    {
        CurrentSource->ops.PutFeature(CurrentSource, featureListView, ActiveViewIndex, ActiveView);
        ActiveViewIndex = 0xFF;
        ActiveView = DS_GetFeature(CurrentSource, featureListView, ActiveViewIndex);
        /* Make sure that the top record stays the same... */
        CurrentRecord = Gridp->ops->GridGetTopIndex(Gridp);
			
        /* Tell the list form to reinitialize the Grid. */
        FrmUpdateForm(formID_ListView, updateChangeViewWithTrigger);
    }
    /* If the user switched views, move to the new view. */
    /* new_view_index != ActiveViewIndex */
	else if ( new_view_index >= 0)
    {
			
        /* Update the in-core variables to the new view. */
        if (ActiveViewIndex == 0xFF)
            DS_PutFeature(CurrentSource, featureListView, ActiveViewIndex, ActiveView);
        else
            CurrentSource->ops.PutFeature(CurrentSource, featureListView, ActiveViewIndex, ActiveView);
        ActiveViewIndex = new_view_index;
        ActiveView = CurrentSource->ops.GetFeature(CurrentSource, featureListView, ActiveViewIndex);
        /* Make sure that the top record stays the same... */
        CurrentRecord = Gridp->ops->GridGetTopIndex(Gridp);
			
        /* Tell the list form to reinitialize the Grid. */
        FrmUpdateForm(formID_ListView, updateChangeViewWithTrigger);
	}

    /* free the view names. */
    for (i = 0; i < numViews + 2; ++i)
    {
        DeallocateMemPtr(view_names[i]);
    }
    DeallocateMemPtr(view_names);

}

static void
DoQuickSort( UInt16 fieldNum , UInt16 sortOrder)
{
    DataSourceSortInfo info;
    DataSourceSortField fields[1];
    void * indicator_context;

    /* Build the information used by the standard sorting callback. */
    info.numFields = 1;
    info.fields = &fields[0];
    info.fields[0].fieldNum = fieldNum;
    info.fields[0].direction = sortOrder;
    info.fields[0].case_sensitive = false;

    /* Tell the data source to sort itself. */
    indicator_context = ShowWorkingIndicator();
    CurrentSource->ops.Sort(CurrentSource, DS_StandardSortCallback, &info);
    HideWorkingIndicator(indicator_context);
    Gridp->ops->GridSetTopIndex(Gridp, CurrentRecord);
	ListViewUpdateForm( FrmGetActiveForm(), updateChangeForm );

}

enum {
        actionID_SortAscend,
        actionID_SortDescend,
        actionID_RunReport
};

static ActionDetails
popup_menu[] = {
    { stringID_SortAscend,              actionID_SortAscend,    false },
    { stringID_SortDescend,     actionID_SortDescend,   false },
#ifdef CONFIG_REPORT
    { stringID_RunReport,   actionID_RunReport, false},
#endif
    { 0,                                0,                      true}
};

static Boolean
ListViewUpdateForm(FormPtr form, UInt16 type)
{
    ControlType * ctl;
    ListType * lst;
    static char viewName[12];
	Int16 i;

    if (type & updateResizeForm) {
        UInt16 dX, dY;

        UtilMoveResizeForm(form, &dX, &dY, true);

        if (dX || dY)
        {
            UtilMoveObject(form, ctlID_ListView_DoneButton, 0, dY, true);
            UtilMoveObject(form, ctlID_ListView_NewButton, 0, dY, true);
            UtilMoveObject(form, ctlID_ListView_FilterTrigger, 0, dY, true);
            UtilMoveObject(form, ctlID_ListView_FilterList, 0, dY, true);
            UtilMoveObject(form, ctlID_ListView_FindButton, 0, dY, true);
            UtilMoveObject(form, ctlID_ListView_RepeatButton, 0, dY, true);
            UtilMoveObject(form, ctlID_ListView_GridCheckbox, 0, dY, true);
            UtilMoveObject(form, ctlID_ListView_ActionList, 0, dY, true);
#ifdef CONFIG_ICONS
#ifdef CONFIG_PLUGINS
            UtilMoveObject(form, ctlID_ListView_PlugInButton, 0, dY, true);
#endif
            UtilMoveObject(form, bmpID_LockButton, 0, dY, true);
#ifdef CONFIG_PALMOS31
            UtilMoveObject(form, bmpID_ExitButton, 0, dY, true);
            UtilMoveObject(form, bmpID_NewRButton, 0, dY, true);
            UtilMoveObject(form, bmpID_FindButton, 0, dY, true);
            UtilMoveObject(form, bmpID_PlugInButton, 0, dY, true);
#endif
#endif
            UtilMoveObject(form, ctlID_ListView_UpButton, dX, dY, true);
            UtilMoveObject(form, ctlID_ListView_DownButton, dX, dY, true);
            UtilMoveObject(form, ctlID_ListView_RightButton, dX, dY, true);
            UtilMoveObject(form, ctlID_ListView_LeftButton, dX, dY, true);
            UtilMoveObject(form, ctlID_ListView_ViewTrigger, dX, 0, true);
            UtilMoveObject(form, ctlID_ListView_ViewList, dX, 0, true);
            UtilResizeObject(form, ctlID_ListView_Grid, dX, dY, true);
        }
    }

/* 
 * We need to do the global script very early on, since filtering
 * may influence the how many records are visible.  We need to know
 * all of this before the grid is created/updated.
 */	
#ifdef CONFIG_GLB_SCRIPTS
	if ( (type & updateRunGlobalScript) ||
		((type & updateInitForm) && LastFormId[0] != formID_SortEditor) ){

		UInt16 j =0;
		DataSourceFeature_GlobalScript *gsd;
		ScriptContext sc;
		CalculatedValue cv;
		UInt8 *bc;

		j = ((ActiveView->flags & VIEWFLAG_SCRIPT_INDEX_MASK) >> 1) & 0x006f;
		if ( j > 0 && j <= CurrentSource->ops.GetNumFeatures( CurrentSource, featureGlobalScript ) ) {
			gsd = CurrentSource->ops.GetFeature( CurrentSource, featureGlobalScript, j-1 ); // -1 since script==0 is no script
			if ( ScriptRequiresAtMost( &(gsd->c), ARRANGING ) ) {

				sc.context = LIST_VIEW_INIT_CONTEXT;
				sc.src = CurrentSource;
				sc.flags = ScriptContext_DataAccess_NFA;
				// no need for sc.record, or data_access

				if ( gsd->c.bytecode ) {
					bc = MemHandleLock( gsd->c.bytecode );
					CurrentSource->ops.ExecuteScript( &sc, bc,  &cv );
					MemPtrUnlock( bc );
				}
			}
			CurrentSource->ops.PutFeature(CurrentSource, featureGlobalScript,
										  j-1, gsd );
		}
	}
#endif

    /* If the model changed, then reinitialize the grid widget. */
    if (type & updateInitForm){
		if ( !RetainGrid )
        {
			// unless we know otherwise, initialize the grid...
			Gridp->ops->GridInit(Gridp, form, ctlID_ListView_Grid,
								 ctlID_ListView_GridCheckbox, ViewModel, ViewModelData);
		}
		else
        {
			
			RetainGrid = false;
		}
    }
    if (type & (updateChangeForm | updateChangeFont) ) {
        Gridp->ops->GridSetHeaderLook(Gridp, DBPrefs.GridDisplay & Display_Header,
                                DBPrefs.GridDisplay & Display_Header_sep,
                                true,
                                DBPrefs.TheBoldFont);
        Gridp->ops->GridSetColumnLook(Gridp, DBPrefs.GridDisplay & Display_Col_sep);
    }
    if (type & (updateChangeView | updateChangeForm | updateChangeFont) ) {    
//  in my copy this is: 
//        Gridp->ops->GridChange(Gridp, form, ViewModel, 0);
        Gridp->ops->GridChange(Gridp, form, ViewModel, ViewModelData);
    }


    Gridp->ops->GridSetTopIndex(Gridp, CurrentRecord);

    ctl = GetObjectPtr(form, ctlID_ListView_ViewTrigger);
    MemMove( viewName, ActiveView->name, 11);
    CtlSetLabel(ctl, viewName);

    ctl = GetObjectPtr(form, ctlID_ListView_FilterTrigger);
    lst = GetObjectPtr(form, ctlID_ListView_FilterList);
    if ( !(DBPrefs.flags & prefFlagUnuseFilters)
            && CurrentSource->ops.Capable(CurrentSource, DS_CAPABLE_LISTVIEW_FILTER)) {
//        if (((ControlPtr)GetObjectPtr(form, ctlID_ListView_RepeatButton))->attr.visible) {

            FrmHideObject(form, FrmGetObjectIndex(form,
                                              ctlID_ListView_RepeatButton));
#ifdef CONFIG_ICONS
#ifdef CONFIG_PALMOS31
            FrmHideObject(form, FrmGetObjectIndex(form,
                                              bmpID_RepeatButton));
#endif
#else
            FrmHideObject(form, FrmGetObjectIndex(form,
                                              ctlID_ListView_FindButton));
						

#endif
            FrmShowObject(form, FrmGetObjectIndex(form,
                                              ctlID_ListView_FilterTrigger));

//        } // Repeat button was visible


#if defined(CONFIG_ICONS) && defined(CONFIG_PLUGINS)
		if ( PlugInMenuItems == NULL ) {
            FrmHideObject(form, FrmGetObjectIndex(form,
											  ctlID_ListView_PlugInButton));

#ifdef CONFIG_PALMOS31
			FrmHideObject( form, FrmGetObjectIndex(form, bmpID_PlugInButton));
#endif			
			
		} else {
			RectangleType r;
			UInt16 x, li;

			li = FrmGetObjectIndex( form, ctlID_ListView_PlugInButton );

            FrmShowObject(form, li );
			
			FrmGetObjectBounds( form, FrmGetObjectIndex(form,
											ctlID_ListView_RepeatButton), &r);
			// the PlugIn button will replace the RepeatFind button when the filters are activated
			x = r.topLeft.x;
			FrmGetObjectBounds( form, li, &r);

			r.topLeft.x = x;			
 			FrmSetObjectBounds( form, li, &r);
#ifdef CONFIG_PALMOS31
			FrmShowObject( form,  FrmGetObjectIndex(form, bmpID_PlugInButton));
			FrmSetObjectBounds( form, FrmGetObjectIndex(form,
												bmpID_PlugInButton), &r);
#endif			
		}
#endif

		i = CurrentSource->ops.GetCurrentFilter( CurrentSource );
		
		// if we're in the dbTempResultsFilter (i == -1), display
		// the an item that is not in the list...
		// otherwise display the filter's label
		if ( i == dbTempResultsFilter ) 
			CtlSetLabel(ctl, "<search>");
		else if ( i >= dbNoFilter )			
			CtlSetLabel( ctl, LstGetSelectionText( lst, i + 1 ) );
		
		
    } else {
        if (CurrentSource->ops.Capable(CurrentSource, DS_CAPABLE_LISTVIEW_FILTER)) {
            CurrentSource->ops.SetFilter( CurrentSource, dbNoFilter, true, NULL);
        }
        FrmHideObject(form, FrmGetObjectIndex(form,
                                              ctlID_ListView_FilterTrigger));
        FrmShowObject(form, FrmGetObjectIndex(form,
                                              ctlID_ListView_RepeatButton));
#if defined(CONFIG_ICONS) && defined(CONFIG_PLUGINS)
		if ( PlugInMenuItems == NULL ) {
            FrmHideObject(form, FrmGetObjectIndex(form,
											  ctlID_ListView_PlugInButton));
			
#ifdef CONFIG_PALMOS31
			FrmHideObject( form, FrmGetObjectIndex(form, bmpID_PlugInButton));
#endif			
			
		}
		else  {

			RectangleType r;
			UInt16 x, li;

			li = FrmGetObjectIndex( form, ctlID_ListView_PlugInButton );
			
			FrmShowObject(form, li );
			FrmGetObjectBounds( form, FrmGetObjectIndex(form,
											ctlID_ListView_RepeatButton), &r);
			
			x = r.topLeft.x + r.extent.x + 2;
			FrmGetObjectBounds( form, li, &r);

			r.topLeft.x = x;			
 			FrmSetObjectBounds( form, li, &r);
#ifdef CONFIG_PALMOS31
			FrmShowObject(form, FrmGetObjectIndex(form,
                                              bmpID_RepeatButton));
			FrmSetObjectBounds( form, FrmGetObjectIndex(form,
											  bmpID_PlugInButton), &r);

#endif
		}
#else
        FrmShowObject(form, FrmGetObjectIndex(form,
                                              ctlID_ListView_FindButton));
#endif
    }        

    /* Redraw the entire display. */
// temporary bug correction for Clie support
	if (Gridp->screenRowSelected != NoneScreenRowSelected) {
		Gridp->ops->GridSetTopIndex(Gridp, Gridp->screenRowSelected);
	}
    FrmDrawForm(FrmGetActiveForm());
    Gridp->ops->GridDraw(Gridp);

    return true;
}

/***
 * function: ListViewHandler
 *
 * arguments:
 *          EventPtr event          the event to treat
 *
 * return:
 *          1                                   for event treated
 *          0                                   for event NOT treated
 *
 * description:
 *          This function is called for all events sended to the ListView form.
 *
 ***/
Boolean
ListViewHandleEvent(EventPtr event)
{
    FormPtr form = NULL;
    ListType * lst;
    UInt16 action;
	Int16 sel;
    static char title[21];


    switch (event->eType) {
    case frmOpenEvent:
    {
        form = FrmGetActiveForm();
        CurrentRecord = CurrentSource->ops.GetOption_UInt16(CurrentSource, optionID_ListView_TopVisibleRecord);
#ifdef CONFIG_PLUGINS
        if (DBPrefs.flags & prefFlagAutoRunPlugin)
        {
            PlugEnvironement pluginInfo;
            pluginInfo.screen = PLUGINTYPE_LISTVIEW;
            pluginInfo.moment = FORM_OPEN;
            pluginInfo.source = CurrentSource;
            pluginInfo.record = CurrentRecord;
            Plug_AutoRun( pluginInfo);
        }
#endif

        /* Get the active list view definition. */
        ActiveViewIndex =
            CurrentSource->ops.GetOption_UInt16(CurrentSource,
                                                optionID_ListView_ActiveView);
        if (ActiveViewIndex == 0xFF)
            ActiveView = DS_GetFeature(CurrentSource, featureListView, ActiveViewIndex);
        else
            ActiveView = CurrentSource->ops.GetFeature(CurrentSource, featureListView, ActiveViewIndex);

		if ( (DBPrefs.flags & prefFlagRRAfterEdit) && 
			 CurrentSource->ops.RecalculateFields ) 
			CurrentSource->ops.RecalculateFields( CurrentSource );
			 
        ViewModel = ViewModel_New( CurrentSource, ActiveView, &ViewModelData);


		FreeFindInfoData( FindInfo );
		FindInfo.type = DSF_FILTER_TYPE_SIMPLE;
		FindInfo.data.simple = AllocateMemPtr( sizeof( SimpleFilterData ) );
		FindInfo.data.simple->caseSensitive = 
			CurrentSource->ops.GetOption_Boolean( CurrentSource,
												  optionID_LFind_CaseSensitive );
		FindInfo.data.simple->wholeWord = 
			CurrentSource->ops.GetOption_Boolean(CurrentSource,
												 optionID_LFind_WholeWord);
		
		FindInfo.data.simple->matchString[0] = '\0';


        /* Make the form title reflect the database name. */
        StrCopy(title, "DB: ");
        CurrentSource->ops.GetTitle(CurrentSource, title+4, sizeof(title)-4);
        if (FntCharsWidth(title, StrLen(title)) < 75)
            title[sizeof(title)-1] = '\0';
        else
            while (FntCharsWidth(title, StrLen(title)) > 75)
            {
                StrCopy( title+StrLen(title)-4, "...");
            }
        FrmSetTitle(form, title);

        /* Setup the action list and trigger. */
        SetupPopupList( form, ctlID_ListView_ActionList, popup_menu);

        /* Hide the "new" button if we are in read-only mode. */
        if (CurrentSource->openMode == dmModeReadOnly) {
#define HideObject(f,id) FrmHideObject((f),FrmGetObjectIndex((f),(id)))
#ifdef CONFIG_ICONS
#ifdef CONFIG_PALMOS31
            HideObject(form, bmpID_NewRButton);
#endif
            FrmShowObject(form, FrmGetObjectIndex(form,
                                                  bmpID_LockButton));
#endif
            HideObject(form, ctlID_ListView_NewButton);
#undef HideObject
        }
		{
			UInt16 i, n;
			DataSourceFeature_Filter *f;

			n = CurrentSource->ops.GetNumFeatures(CurrentSource, featureFilter );
			FilterMenuItems = AllocateMemPtr( sizeof( char * ) * (n+2));
			FilterMenuItems[0] = AllocateMemPtr( 10 );
			StrCopy( FilterMenuItems[0], "No Filter" );
			FilterMenuItems[n+1] = AllocateMemPtr( 8 );
			StrCopy( FilterMenuItems[n+1], "Edit..." );
			for(i = 1; i <= n; i++) {
				FilterMenuItems[i] = AllocateMemPtr( 32 );
				f = CurrentSource->ops.GetFeature( CurrentSource, featureFilter, i-1 );
				StrCopy( FilterMenuItems[i], f->name );
				CurrentSource->ops.PutFeature(CurrentSource, featureFilter, 
											  i-1, f );
			}
			
			lst = GetObjectPtr(form, ctlID_ListView_FilterList);
			LstSetListChoices(lst, FilterMenuItems, n+2);
        }
#if defined(CONFIG_PLUGINS) && defined(CONFIG_ICONS)
        {
			UInt16 i;

			for( i = 0; i < MAXPLUGIN; i++ ) {
				if ( !PluginTable[i].dbRef ) break;
			}
			if ( i == 0 )
                PlugInMenuItems = 0;
			else {
                UInt16 n;
                RectangleType r;

				PlugInMenuItems = AllocateMemPtr( sizeof( char * ) * i );
				for( i = 0; i < MAXPLUGIN; i++ ) {
					if ( PluginTable[i].dbRef ) {
						PlugInMenuItems[i] = PluginTable[i].name;
					}
					else break;
				}

				/* Set the items in the PlugIn Menu, then resize it.
				 * finally, relocate it, keeping its bottom edge in the
				 * original position
				 */
				lst = GetObjectPtr(form, ctlID_ListView_PlugInList);
				LstSetListChoices(lst, PlugInMenuItems, i );
				LstSetHeight( lst, i );


				FrmGetObjectBounds( form, FrmGetObjectIndex( form, ctlID_ListView_PlugInButton ), &r );
				i = r.topLeft.y + r.extent.y;
				n = FrmGetObjectIndex( form, ctlID_ListView_PlugInList );
				FrmGetObjectBounds( form, n, &r );
				r.topLeft.y = i - r.extent.y;
				FrmSetObjectBounds( form, n, &r );


			}
        }
#endif

        Gridp = Grid_New();
        ListViewUpdateForm(form, updateInitForm | updateResizeForm);

        return true;
    }
    case frmCloseEvent:
	{
		Int16 n, i;


#ifdef CONFIG_PLUGINS
        if (DBPrefs.flags & prefFlagAutoRunPlugin)
        {
            PlugEnvironement pluginInfo;
            pluginInfo.screen = PLUGINTYPE_LISTVIEW;
            pluginInfo.moment = FORM_CLOSE;
            pluginInfo.source = CurrentSource;
            pluginInfo.record = CurrentRecord;
            Plug_AutoRun( pluginInfo);
        }
#endif

        CurrentSource->ops.SetOption_UInt16(CurrentSource,
                                            optionID_ListView_TopVisibleRecord,
                                            Gridp->ops->GridGetTopIndex(Gridp));

		/**
		 * BUG.
		 *   I don't quite understand it, but when the database is
		 *    over the AUTO_RETAIN_SIZE, you can make db crash by:
		 *   1) enter the db
		 *   2) create a new record
		 *   3) press the Application Launcher button
		 *
		 *   according to the debugger, the crash occurs in 
		 *  FrmCloseAllForms() when a freed ptr or handle is accessed
		 *  if we remove the Gridp->ops->GridFree(Gridp) call below
		 *  this bug does not occur.
		 *
		 *  I think that it might be related to the offscreen drawing
		 *  window...
		 */
		if ( !RetainGrid  ) {
			Gridp->ops->GridFree(Gridp);
		}

        ViewModel_Delete( CurrentSource, ActiveView, &ViewModel, &ViewModelData);
		/**
		 * BUG?
		 * Do we really need this if statement?
		 * it should be ok just to call CurrentSource->ops.PutFeature()
		 */
        if (ActiveViewIndex == 0xFF)
            DS_PutFeature(CurrentSource, featureListView, ActiveViewIndex, ActiveView);
        else
            CurrentSource->ops.PutFeature(CurrentSource, featureListView,  ActiveViewIndex, ActiveView);
        CurrentSource->ops.SetOption_UInt16(CurrentSource,
                                            optionID_ListView_ActiveView,
                                            ActiveViewIndex);

		/* Free the filter list */
		if ( FilterMenuItems != NULL ) {
			lst = GetObjectPtr(FrmGetActiveForm(), ctlID_ListView_FilterList);
			n = LstGetNumberOfItems( lst );
			
			for( i = 0; i < n; i++ ) {
				DeallocateMemPtr( FilterMenuItems[i] );
			}
			DeallocateMemPtr(FilterMenuItems);
			FilterMenuItems = NULL;
		}
#if defined(CONFIG_PLUGINS) && defined(CONFIG_ICONS)
		if ( PlugInMenuItems != NULL ) {
			// the menu items are just pointers to strings allocated elsewhere
			DeallocateMemPtr( PlugInMenuItems );
			PlugInMenuItems = NULL;
		}
#endif
        FreePopupList( popup_menu);

	}
	break;

    case winDisplayChangedEvent:
        form = FrmGetActiveForm();
        return ListViewUpdateForm(form, updateResizeForm | updateInitForm);

    case frmUpdateEvent:
        form = FrmGetActiveForm();
        return ListViewUpdateForm(form, event->data.frmUpdate.updateCode);

    case frmGotoEvent:
        form = FrmGetActiveForm();
        Gridp->ops->GridSetTopIndex(Gridp, event->data.frmGoto.recordNum);
        Gridp->ops->GridDraw(Gridp);
        return true;

    case penDownEvent:
        if (Gridp->ops->GridHandleEvent(Gridp, event))
            return true;
        break;

    case gridSelectEvent:
      {
#ifndef CONFIG_VLDB
        UInt32 numRecords;

        if (CurrentSource->ops.Capable(CurrentSource, DS_CAPABLE_LISTVIEW_FILTER)) {
            UInt16 currentFilter = CurrentSource->ops.GetCurrentFilter(CurrentSource);
            CurrentSource->ops.SetFilter( CurrentSource, dbNoFilter, false, NULL);
            CurrentSource->ops.Info(CurrentSource, NULL, NULL, NULL, NULL, &numRecords, NULL);
            CurrentSource->ops.SetFilter( CurrentSource, currentFilter, false, &FindInfo);
        } else {
            CurrentSource->ops.Info(CurrentSource, NULL, NULL, NULL, NULL, &numRecords, NULL);
        }

        if (event->data.tblSelect.row < numRecords)
#else


		if (event->data.tblSelect.row <= Gridp->lastModelIndex )
#endif
        {
            CurrentRecord = event->data.tblSelect.row;
            IsNewRecord = false;
			// we only keep the cache around for reasonably large databases
			// for small ones, (with number of records < screen size)
			// it is important to rebuild...
			// see the note in grid.c regarding grid->numModelRows
			if ( Gridp->numModelRows > AUTO_RETAIN_SIZE )
				RetainGrid = true;
            FrmGotoForm(formID_EditView);
        }
        return true;
      }
    case gridHeaderSelectEvent:
        if (CurrentSource->openMode != dmModeReadOnly) {
            form = FrmGetActiveForm();
            action = ShowPopupList( form, ctlID_ListView_ActionList, event);

            switch( action) {
            case actionID_SortAscend:
                if (CurrentSource->ops.Sort) {
                    DoQuickSort( ActiveView->cols[event->data.tblSelect.column].fieldNum,
                                        SORT_DIRECTION_ASCENDING);
                    Gridp->ops->GridDraw(Gridp);
                }
            break;
            case actionID_SortDescend:
                if (CurrentSource->ops.Sort) {
                    DoQuickSort( ActiveView->cols[event->data.tblSelect.column].fieldNum,
                                        SORT_DIRECTION_DESCENDING);
                    Gridp->ops->GridDraw(Gridp);
                }
            break;
#ifdef CONFIG_REPORT
            case actionID_RunReport:
			{
				NoteType note;

				StrCopy( note.title, "Report" );

				note.h = MemHandleNew( 256 );
				printReport( CurrentSource, ActiveView->cols[event->data.tblSelect.column].fieldNum, MemHandleLock( note.h ) );

				NotePopupForm( &note, false );

			}
			break;
#endif
			default:
            }
        }
    return true;

    case gridVerticalScrollersEvent:
        UpdateScrollers(FrmGetActiveForm(),
                        event->data.generic.datum[1],
                        event->data.generic.datum[2]);
        return true;

    case gridHorizontalScrollersEvent:
        UpdateFieldButtons( FrmGetActiveForm(),
                            event->data.generic.datum[1],
                            event->data.generic.datum[2]);
        return true;

#ifdef CONFIG_HANDERA
    case menuOpenEvent:
        if (FeatureSet.Handera) {
            MenuAddItem(menuitemID_AppPrefs, menuitemID_HanderaRotate,
                                '\0', RotateString);
        }
        break;
#endif

    case menuCmdBarOpenEvent:
        if (FeatureSet.OS35) {
#ifdef CONFIG_ICONS
            MenuCmdBarAddButton(menuCmdBarOnRight, bmpID_GotoTop, menuCmdBarResultMenuItem, menuitemID_GotoTop, 0);
            MenuCmdBarAddButton(menuCmdBarOnRight, bmpID_GotoBottom, menuCmdBarResultMenuItem, menuitemID_GotoBottom, 0);
#endif
            MenuCmdBarAddButton(menuCmdBarOnRight, BarDeleteBitmap, menuCmdBarResultMenuItem, menuitemID_DeleteAllRecords, 0);
        }
        return false;

    case menuEvent:
        return DoCommand(event->data.menu.itemID);

    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_ListView_DoneButton:
            FrmGotoForm(formID_Chooser);
            return true;

        case ctlID_ListView_NewButton:
            IsNewRecord = true;
			Gridp->numModelRows++;
			// we only keep the cache around for reasonably large databases
			// for small ones, (with number of records < screen size)
			// it is important to rebuild...
			// see the note in grid.c regarding grid->numModelRows
			if ( Gridp->numModelRows > AUTO_RETAIN_SIZE )
				RetainGrid = true;
            FrmGotoForm(formID_EditView);
            return true;

        case ctlID_ListView_FindButton:
			FindInfo.forFilterDefinition = -1;
            FrmPopupForm( formID_FindDialog);
            return true;
			
        case ctlID_ListView_RepeatButton:
		{
			UInt16 frec;

			FindInfo.forFilterDefinition = -1;
            if ( CurrentSource->ops.Search != 0 )
				frec = CurrentSource->ops.Search( CurrentSource, &FindInfo,
												  dbNoFilter, FILTER_INTERSECTION);
			else 
				frec = DS_Search( CurrentSource, &FindInfo,
								  dbNoFilter, FILTER_INTERSECTION);
			
			ListViewUpdateAfterFind( dbNoFilter, frec, &FindInfo );
			return true;
		}

#if defined(CONFIG_ICONS) && defined(CONFIG_PLUGINS)
		case ctlID_ListView_PlugInButton:

			if ( PlugInMenuItems == NULL ) {
				DbgPrintF( "No PlugIns Installed!" );
			}
			else {
				form = FrmGetActiveForm();
				lst = GetObjectPtr(form, ctlID_ListView_PlugInList );
				LstSetSelection( lst, noListSelection );

				action = LstPopupList( lst );
				if ( action != noListSelection ) {
					PlugEnvironement e;
					e.screen = PLUGINTYPE_LISTVIEW;
					e.source = CurrentSource;
					Plug_Execute( action, e, selector );
				}
			}
			break;
#endif

        case ctlID_ListView_FilterTrigger:
		{
			UInt16 numUserFilters;

            if( (DBPrefs.flags & prefFlagUnuseFilters)
                    || !CurrentSource->ops.Capable(CurrentSource, DS_CAPABLE_LISTVIEW_FILTER))
                return true;
            form = FrmGetActiveForm();
            lst = GetObjectPtr(form, ctlID_ListView_FilterList);
            sel = CurrentSource->ops.GetCurrentFilter(CurrentSource);
			sel++; // script index 0 is menu item 1
			if ( sel < 0 ) LstSetSelection( lst, noListSelection );
			else LstSetSelection( lst, sel );

			action = LstPopupList(lst);

			numUserFilters = CurrentSource->ops.GetNumFeatures( CurrentSource, featureFilter );
			
			if ( action != noListSelection ) {

				if ( action > numUserFilters ) 
					action = (action - numUserFilters) + 0x0f00;
				
				switch( action ){
				case 0:
					CurrentSource->ops.SetFilter(CurrentSource, dbNoFilter, true, NULL);
					break;
					
				case 0x0f01:  // 'edit filters...' item
					FeatureEditor_Feature = featureFilter;
					FrmGotoForm(formID_NamesEditor);
					return true;
					
				default:
					
					CurrentSource->ops.SetFilter(CurrentSource, action-1, true, &FindInfo);
					break;
				}

				FrmUpdateForm(formID_ListView, updateChangeView);
			}
            return true;
		}
        case ctlID_ListView_ViewTrigger:
            SelectViewTrigger();
            return true;
        }
        break;

    case ctlRepeatEvent:
        switch (event->data.ctlRepeat.controlID) {
        case ctlID_ListView_UpButton:
            if (DBPrefs.flags & prefFlagUpArrowPage)
                Gridp->ops->GridScrollVertical(Gridp, winUp, Gridp->ops->GridGetNumberOfRecordsInPage(Gridp));
            else
                Gridp->ops->GridScrollVertical(Gridp, winUp, 1);
            break;

        case ctlID_ListView_DownButton:
            if (DBPrefs.flags & prefFlagDownArrowPage)
                Gridp->ops->GridScrollVertical(Gridp, winDown, Gridp->ops->GridGetNumberOfRecordsInPage(Gridp));
            else
                Gridp->ops->GridScrollVertical(Gridp, winDown, 1);
            break;

        case ctlID_ListView_LeftButton:
            Gridp->ops->GridScrollHorizontal(Gridp, winLeft, 1);
            Gridp->ops->GridDraw(Gridp);
            break;

        case ctlID_ListView_RightButton:
            Gridp->ops->GridScrollHorizontal(Gridp, winRight, 1);
            Gridp->ops->GridDraw(Gridp);
            break;
        }
        break;

    case keyDownEvent:
        switch (event->data.keyDown.chr) {
		case vchrPalmMin + 5:
			PINSetInputAreaState(pinInputAreaClosed);
			return true;
#ifdef CONFIG_JOGDIAL
        case vchrRockerLeft:
            return ListViewMove(winLeft);

        case vchrRockerRight:
            return ListViewMove(winRight);

        case vchrRockerUp:
            return ListViewMove(winUp);

        case vchrRockerDown:
            return ListViewMove(winDown);

        case vchrRockerCenter:
            return ListViewMove(-1);

#ifdef CONFIG_HANDERA
        case chrEscape:
#endif
#ifdef CONFIG_SONY
        case vchrJogBack:
#endif
            if (Gridp->screenRowSelected == NoneScreenRowSelected) {
                Gridp->ops->GridUnhighlightAll(Gridp);
                Gridp->ops->GridDraw(Gridp);
            } else {
                FrmGotoForm(formID_Chooser);
            }
            return true;
#endif /*#ifdef CONFIG_JOGDIAL*/

        default:
            if (event->data.keyDown.chr < 0x7F && TxtGlueCharIsAlNum(event->data.keyDown.chr)){
                CurrentRecord = LookUpDataSource(CurrentSource, ActiveView->cols[0].fieldNum, event->data.keyDown.chr);
#ifdef CONFIG_JOGDIAL
                Gridp->ops->GridSelectRecordIndex(Gridp, CurrentRecord);
                Gridp->ops->GridSelectRecordIndex(Gridp, CurrentRecord);
#endif
                FrmUpdateForm(formID_ListView, updateRedrawForm);
            }
        }
        break;

    default:
        break;
	}

    return false;
}


void
ListViewUpdateAfterFind(Int16 filterId, UInt16 foundRecNum, FindData *findInfo) {
    /* Highlight the found record or scroll grid so record is visible. */

	Gridp->ops->GridUnhighlightAll(Gridp);
    if (filterId != dbNoFilter) {
        if (foundRecNum != noRecord) {
//            CurrentSource->ops.SetFilter( CurrentSource, filter, false);
            CurrentRecord = foundRecNum;
        } else {
            FrmAlert(alertID_MatchNotFound);
            CurrentSource->ops.SetFilter( CurrentSource, dbNoFilter, true, NULL );
            CurrentRecord = 0;
        }
        FrmUpdateForm(formID_ListView, updateChangeView);
    } else {
        if (foundRecNum != noRecord) {
            Gridp->ops->GridSetTopIndex(Gridp, foundRecNum);
#ifdef CONFIG_JOGDIAL
            Gridp->ops->GridSelectRecordIndex(Gridp, foundRecNum);
#else
            Gridp->ops->GridHighlightIndex(Gridp, foundRecNum);
#endif
            CurrentRecord = foundRecNum;
            findInfo->lastFound = foundRecNum;
        } else {
            FrmAlert(alertID_MatchNotFound);
            Gridp->ops->GridSetTopIndex(Gridp, 0);
            Gridp->ops->GridUnhighlightAll(Gridp);
#ifdef CONFIG_JOGDIAL
            Gridp->ops->GridSelectRecordIndex(Gridp, dmMaxRecordIndex);
#else
            Gridp->ops->GridUnhighlightAll(Gridp);
#endif
            CurrentRecord = 0;
            findInfo->lastFound = 0xFFFF;
        }
        FrmUpdateForm(formID_ListView, updateRedrawForm);
    }
}

static Boolean
ListViewMove(WinDirectionType direction)
{
	Boolean ret = false;
    switch (direction) {
    case winUp:
        if (Gridp->screenRowSelected != NoneScreenRowSelected) {
            if (Gridp->screenRowSelected == Gridp->ops->GridGetTopIndex(Gridp))
                Gridp->ops->GridScrollVertical(Gridp, winUp, 1);
            ret = Gridp->ops->GridMoveSelectedLine(Gridp, winUp);
        } else {
            if (DBPrefs.flags & prefFlagPageUpPage)
                ret = Gridp->ops->GridScrollVertical(Gridp, winUp, Gridp->ops->GridGetNumberOfRecordsInPage(Gridp));
            else
                ret = Gridp->ops->GridScrollVertical(Gridp, winUp, 1);
		}
    break;
    case winDown:
        if (Gridp->screenRowSelected != NoneScreenRowSelected) {
            if ( (Gridp->screenRowSelected - Gridp->ops->GridGetTopIndex(Gridp)) ==
				 (Gridp->ops->GridGetNumberOfRecordsInPage(Gridp) - 1))
                Gridp->ops->GridScrollVertical(Gridp, winDown, 1);
            ret = Gridp->ops->GridMoveSelectedLine(Gridp, winDown);
        } else {
            if (DBPrefs.flags & prefFlagPageDownPage)
                ret = Gridp->ops->GridScrollVertical(Gridp, winDown, Gridp->ops->GridGetNumberOfRecordsInPage(Gridp));
            else
                ret = Gridp->ops->GridScrollVertical(Gridp, winDown, 1);
        }
    break;
    case winLeft:
            ret = Gridp->ops->GridScrollHorizontal(Gridp, winLeft, 1);
            Gridp->ops->GridDraw(Gridp);
    break;
    case winRight:
            ret = Gridp->ops->GridScrollHorizontal(Gridp, winRight, 1);
            Gridp->ops->GridDraw(Gridp);
    break;
    default:
        if (Gridp->screenRowSelected == NoneScreenRowSelected) {
            CurrentRecord =  Gridp->screenRowSelected = Gridp->ops->GridGetTopIndex(Gridp);
            ret = Gridp->ops->GridMoveSelectedLine(Gridp, -1);
	    } else {
            CurrentRecord =  Gridp->screenRowSelected;
            if (CurrentSource->ops.Seek(CurrentSource, &CurrentRecord, 0, dmSeekForward))
            {
                IsNewRecord = false;
                FrmGotoForm(formID_EditView);
            }
			ret = true;
        }   
    break;
    }
	return ret;
}
