/*
 * DB: Display model object
 * Copyright (C) 1990-2001: Tom Dyas (original code from differents other files)
 * Copyright (C) 2001-2003: Marc Chalain (marc-chalain@voila.fr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <PalmOS.h>

#include "db.h"
#include "grid.h"
#include "io.h"

static DataSourceGetter Getter;
static DataSourceFeature_ListView * ActiveView;
static DataSource * Source;

static UInt16 ViewModel_GetColumnCount(void * __unused) SECT_UI;
static void ViewModel_GetColumnName(void *, UInt16, Char *, UInt32) SECT_UI;
static UInt16 ViewModel_GetColumnWidth(void *, UInt16) SECT_UI;
static void ViewModel_SetColumnWidth(void *, UInt16, UInt16) SECT_UI;
static void * ViewModel_LockIndex(void *, UInt16, Boolean) SECT_UI;
static void ViewModel_UnlockIndex(void *, UInt16, Boolean, void *) SECT_UI;
static GridCellType ViewModel_GetCellType(void *,UInt16,UInt16) SECT_UI;
static void ViewModel_GetCellString(void *, UInt16, UInt16, Char *,
                                    UInt32, void *) SECT_UI;
static Boolean ViewModel_GetCellBoolean(void *, UInt16, UInt16, void *_data) SECT_UI;
static Boolean bool_getter_GetBoolean(void * data, UInt16 fieldNum) SECT_UI;
static void ViewModel_SetCellBoolean(void *, UInt16, UInt16, Boolean,
                                     void *) SECT_UI;
static void ViewModel_GetCellID(void *, UInt16, UInt32 *) SECT_UI;
static Boolean ViewModel_Seek(void *, UInt16 *, Int16, Int16) SECT_UI;
static UInt16 ViewModel_NumRows(void *) SECT_UI;
static Boolean ViewModel_IsColumnEditable(void * __unused, UInt16 col) SECT_UI;

static GridModelType ViewModel = {
    GetColumnCount: ViewModel_GetColumnCount,
    GetColumnName: ViewModel_GetColumnName,
    GetColumnWidth: ViewModel_GetColumnWidth,
    SetColumnWidth: ViewModel_SetColumnWidth,
    LockIndex: ViewModel_LockIndex,
    UnlockIndex: ViewModel_UnlockIndex,
    GetCellType: ViewModel_GetCellType,
    GetCellString: ViewModel_GetCellString,
    GetCellBoolean: ViewModel_GetCellBoolean,
    SetCellBoolean: ViewModel_SetCellBoolean,
    GetCellID: ViewModel_GetCellID,
    Seek: ViewModel_Seek,
    NumRows: ViewModel_NumRows,
    IsColumnEditable: ViewModel_IsColumnEditable
};

GridModelType * ViewModel_New( DataSource * source, DataSourceFeature_ListView *activeView, void ** model_data)
{
    ActiveView = activeView;
    Source = source;
    *model_data = NULL;
    return &ViewModel;
}

void ViewModel_Delete( DataSource * source, DataSourceFeature_ListView *activeView, GridModelType **model, void ** model_data)
{
    *model = NULL;
    *model_data = NULL;
}

static UInt16 ViewModel_GetColumnCount(void * __unused)
{
    return ActiveView->numCols;
}

static void ViewModel_GetColumnName(void * __unused, UInt16 col,
                                    Char * name, UInt32 len)
{
    StrNCopy(name,
             Source->schema.fields[ActiveView->cols[col].fieldNum].name,
             len);
}

static UInt16 ViewModel_GetColumnWidth(void * __unused, UInt16 col)
{
    UInt16 value = ActiveView->cols[col].width;
#ifdef CONFIG_HANDERA
    if (FeatureSet.Handera) {
        value *= 1.5;
    }
#endif
    return value;
}

static void ViewModel_SetColumnWidth(void * __unused, UInt16 col, UInt16 width)
{
}

static void * ViewModel_LockIndex(void * __unused,
                                  UInt16 index, Boolean writable)
{
    void * row_data;

    Source->ops.LockRecord(Source, index, writable,
                                  &Getter, &row_data);
    return row_data;
}

static void ViewModel_UnlockIndex(void * __unused, UInt16 index,
                                  Boolean writable, void * row_data)
{
    Source->ops.UnlockRecord(Source, index, writable, row_data);
}

static GridCellType ViewModel_GetCellType(void * __unused,
                                          UInt16 index, UInt16 col)
{
    switch (Source->schema.fields[ActiveView->cols[col].fieldNum].type) {
    case FIELD_TYPE_BOOLEAN:
        return gridCellTypeBoolean;
    case FIELD_TYPE_NOTE:
        return gridCellTypeNote;
    case FIELD_TYPE_FLOAT:
      return gridCellTypeRJString;
    case FIELD_TYPE_STRING:
    case FIELD_TYPE_INTEGER:
    case FIELD_TYPE_LINK:
    case FIELD_TYPE_LINKED:
    default:
        return gridCellTypeString;
    }
}

static void ViewModel_GetCellString(void * __unused, UInt16 index, UInt16 col,
                                    Char * data, UInt32 len, void * row_data)
{
    const UInt16 fieldNum = ActiveView->cols[col].fieldNum;

    ConvertFieldToString( Source, &Getter, row_data,
                        fieldNum, data, len);
}

static Boolean ViewModel_GetCellBoolean(void * __unused, UInt16 index,
                                        UInt16 col, void * row_data)
{
    return Getter.GetBoolean(row_data, ActiveView->cols[col].fieldNum);
}

static Boolean bool_getter_GetBoolean(void * data, UInt16 fieldNum)
{
    return (data != 0) ? true : false;
}

static DataSourceGetter bool_getter = {
    GetBoolean: bool_getter_GetBoolean,
};

static void ViewModel_SetCellBoolean(void * __unused, UInt16 index,
                                     UInt16 col, Boolean b, void * row_data)
{
    Source->ops.SetField(Source, row_data,
                                ActiveView->cols[col].fieldNum, &bool_getter,
                                b ? ((void *) 1) : ((void *) 0));
}

static void ViewModel_GetCellID(void * __unused, UInt16 index, UInt32 * idP)
{
    UInt32 uniqueID;

    uniqueID = Source->ops.GetRecordID(Source, index);
    *idP = uniqueID;
}

static Boolean ViewModel_Seek(void * __unused, UInt16 * indexP,
                              Int16 offset, Int16 direction)
{
    return Source->ops.Seek(Source, indexP, offset, direction);
}

static Boolean ViewModel_IsColumnEditable(void * __unused, UInt16 col)
{
    Boolean result = false;

    if ((Source->schema.fields[ActiveView->cols[col].fieldNum].type) == FIELD_TYPE_BOOLEAN
            && (CurrentSource->openMode == dmModeReadWrite))
        result = true;

    return result;
}

static UInt16 ViewModel_NumRows(void * __unused)
{
    UInt32 numRecords;

    Source->ops.Info(Source, NULL, NULL, NULL, NULL, &numRecords, NULL);
    return (UInt16)numRecords;
}

